# -*- coding: utf-8 -*-
"""
    (c) 2017 FriskbyBergen.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from friskby_controlpanel import friskby_controlpanel as cp
from tests.fake_friskby_interface import FakeFriskbyInterface
from datetime import datetime
import unittest


class DashboardTestCase(unittest.TestCase):

    def setUp(self):
        self.iface = FakeFriskbyInterface()
        cp.app.config['TESTING'] = True
        cp.app.config['FRISKBY_INTERFACE'] = self.iface
        self.app = cp.app.test_client()

    def test_when_registered_device(self):
        self.iface.device_id = 'mydevice'
        out = self.app.get('/')
        self.assertIn("mydevice", out.data)

    def test_no_sampling_nor_any_upload_has_happened(self):
        self.iface.device_id = 'mydevice'
        out = self.app.get('/')
        self.assertIn(
            """This device is neither sampling, nor uploading anything to the server.""",  # noqa
            out.data
        )

    def test_sampling_but_no_uploading(self):
        self.iface.device_id = 'mydevice'
        self.iface.all_samples_count = 1
        self.iface.most_recently_sampled = datetime(2007, 12, 5, 12, 0, 0)
        out = self.app.get('/')
        self.assertIn(
            """This device is sampling, but has not yet uploaded data to the server. The last sample was taken at 2007-12-05 12:00:00.""",  # noqa
            out.data
        )

    def test_sampling_and_uploading(self):
        self.iface.device_id = 'mydevice'
        self.iface.uploaded_samples_count = 1
        self.iface.all_samples_count = 1
        self.iface.most_recently_uploaded = datetime(2007, 12, 5, 12, 0, 0)
        out = self.app.get('/')
        self.assertIn(
            """This device is sampling, and has uploaded data to the server at 2007-12-05 12:00:00.""",  # noqa
            out.data
        )


    def test_socket_iface_name(self):
        self.iface.device_id = 'mydevice'
        self.assertIn('Could not obtain', self.app.get('/').data)
        self.iface.socket = '27.182.81.82'
        self.assertIn('27.182.81.82', self.app.get('/').data)

    def test_get_recent_samples(self):
        self.iface.device_id = 'mydevice'
        self.assertIn('PM10', self.app.get('/').data)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
    Friskby Controlpanel

    A webapp that allows for configuration of a friskby system on a device.

    (c) 2017 FriskbyBergen.
    (c) 2015 by Armin Ronacher.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
import sys
import inspect
from . import ctljson
from flask import (Flask, request, redirect, flash, url_for, render_template)  # noqa
from .friskby_interface import FriskbyInterface
from .forms import SettingsForm
from .friskby_settings import DEVICE_CONFIG_PATH, config_url


app = Flask(__name__)
app.config.from_object(__name__)

app.config.update(dict(
    FRISKBY_DEVICE_CONFIG_PATH=DEVICE_CONFIG_PATH,
    FRISKBY_INTERFACE=FriskbyInterface,
    CONFIG_URL=config_url,
    WTF_CSRF_ENABLED=False
))
app.config.from_envvar('FRISKBY_CONTROLPANEL_SETTINGS', silent=True)
app.secret_key = 'we actually do not care too much'


@app.before_request
def before_request():
    """Initializes the friskby interface if it wasn't already."""
    iface = app.config['FRISKBY_INTERFACE']
    if inspect.isclass(iface):
        app.config['FRISKBY_INTERFACE'] = iface()


@app.context_processor
def inject_device_id():
    """Injects the device_id into the template context, or None if none."""
    filename = app.config['FRISKBY_DEVICE_CONFIG_PATH']
    try:
        device_id = app.config['FRISKBY_INTERFACE'].get_device_id(filename)
    except IOError:
        device_id = None
    return dict(device_id=device_id)


@app.context_processor
def inject_statuses():
    """Injects the friskby system service statuses into the template context.
    """
    fby_iface = app.config['FRISKBY_INTERFACE']
    sampler_status = fby_iface.get_service_status('sampler')
    submitter_status = fby_iface.get_service_status('submitter')
    friskby_status = fby_iface.get_service_status('friskby')
    friskby_controlpanel_status = fby_iface.get_service_status(
        'friskby_controlpanel'
    )
    return {
        'sampler_status': sampler_status,
        'submitter_status': submitter_status,
        'friskby_status': friskby_status,
        'friskby_controlpanel_status': friskby_controlpanel_status,
    }


@app.route('/')
def dashboard():
    """Renders the dashboard. Will redirect to device registration if no
    device_id was found."""
    fby_iface = app.config['FRISKBY_INTERFACE']
    config_path = app.config['FRISKBY_DEVICE_CONFIG_PATH']

    try:
        device_id = fby_iface.get_device_id(config_path)
    except IOError:
        device_id = None

    # No device, so redirect to the register page.
    if not device_id:
        return redirect(url_for('register'))

    sockname = fby_iface.get_socket_iface_name()
    return render_template(
        'dashboard.html',
        has_sampled=fby_iface.get_all_samples_count() > 0,
        has_uploaded=fby_iface.get_uploaded_samples_count() > 0,
        most_recent_sample=fby_iface.get_most_recently_sampled(),
        most_recent_upload=fby_iface.get_most_recently_uploaded(),
        get_recent_samples=fby_iface.get_recent_samples(),
        socket_iface_address=sockname)


@app.route('/register', methods=['GET', 'POST'])
def register():
    """Renders a page whereon you can register a device."""
    fby_iface = app.config['FRISKBY_INTERFACE']
    config_path = app.config['FRISKBY_DEVICE_CONFIG_PATH']
    error = None

    try:
        device_id = fby_iface.get_device_id(config_path)
    except IOError:
        device_id = None

    if request.method == 'POST' and device_id:
        return redirect(url_for('dashboard'))
    elif request.method == 'POST' and not device_id:
        if request.form['deviceid'] == "":
            error = 'No device id'
        else:
            device_id = request.form['deviceid']
            url = app.config['CONFIG_URL'](device_id)
            print("Fetching config for device %s from: %s" % (device_id, url))
            sys.stdout.flush()
            try:
                fby_iface.download_and_save_config(url, config_path)
                return redirect(url_for('registered'))
            except ValueError as e:
                error = "Failed to download configuration: %s" % e
                print(error)
                sys.stdout.flush()

    return render_template('register.html', error=error)


@app.route('/registered')
def registered():
    """Renders a “your device is registered template”."""
    return render_template('registered.html')


@app.route('/service/<string:service_name>')
def status(service_name):
    """Renders a service given a service_name."""
    iface = app.config['FRISKBY_INTERFACE']
    error = None
    service_status = None
    service_journal = []

    try:
        service_status = iface.get_service_status(service_name)
        service_journal = iface.get_service_journal(service_name)
    except KeyError as e:
        error = 'No such service: %s (%s).' % (service_name, e)
        print(error)
        sys.stdout.flush()
        flash(error)
        return redirect(url_for('dashboard'))
    else:  # There could be something in the journal at this point.
        for key in ctljson.KEYS:
            for entry in service_journal:
                if key in entry:
                    export_key = ctljson.KEYS[key]
                    entry[export_key] = ctljson.parse_message(
                        export_key, entry[key]
                    )

    return render_template(
        'service.html',
        error=error,
        name=service_name,
        status=service_status,
        journal=service_journal
    )


@app.route('/service/<string:service_name>/<string:action_name>',
           methods=['POST'])
def status_manage(service_name, action_name):
    """Manages a service."""
    iface = app.config['FRISKBY_INTERFACE']

    try:
        iface.manage_service(service_name, action_name)
    except KeyError:
        error = 'No such service (%s) or action (%s).' % (service_name,
                                                          action_name)
        print(error)
        sys.stdout.flush()
        flash(error)
        return redirect(url_for('dashboard'))

    flash('Requested %s on service %s.' % (action_name, service_name))
    return redirect(url_for('status', service_name=service_name))


@app.route('/settings', methods=['GET', 'POST'])
def settings():
    """Displays and allows changing of settings."""
    iface = app.config['FRISKBY_INTERFACE']
    form = None

    if request.method == 'GET':
        form = SettingsForm(data=iface.get_settings())
    else:
        form = SettingsForm()  # defaults to flask.request.form
        if form.validate_on_submit():
            flash('Settings saved.')
            iface.set_settings(form.data)
        else:
            flash('Form had errors.')

    return render_template(
        'settings.html', form=form, errors=form.errors
    )



# -*- coding: utf-8 -*-
"""
    (c) 2017 FriskbyBergen.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from friskby_controlpanel import friskby_controlpanel as cp
from tests.fake_friskby_interface import FakeFriskbyInterface
import unittest


class SettingsTestCase(unittest.TestCase):

    def setUp(self):
        self.iface = FakeFriskbyInterface()
        cp.app.config['TESTING'] = True
        cp.app.config['FRISKBY_INTERFACE'] = self.iface

        # Set device so that the dashboard does not redirect.
        self.iface.device_id = 'mydevice'

        self.app = cp.app.test_client()

    def test_looking_at_settings(self):
        out = self.app.get('/settings')
        self.assertIn('rpi_sample_time', out.data)
        self.assertIn('rpi_control_panel_host', out.data)
        self.assertIn('rpi_control_panel_port', out.data)
        self.assertIn('rpi_sds011', out.data)

    def test_changing_a_setting_successfully(self):
        out = self.app.post('/settings', data={
            'rpi_sample_time': 12,
            'rpi_control_panel_host': '0.0.0.0',
            'rpi_control_panel_port': 50,
            'rpi_sds011': '/dev/foo'
        })
        self.assertIn('Settings saved.', out.data)
        self.assertEqual(12, self.iface.settings['rpi_sample_time'])

    def test_posting_form_with_bad_data(self):
        out = self.app.post('/settings', data={
            'rpi_control_panel_host': '0.0.0.0',
            'rpi_control_panel_port': 50,
            'rpi_sds011': '/dev/foo'
        })
        self.assertIn('Form had errors', out.data)
        self.assertIn('This field is required', out.data)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
    (c) 2017 FriskbyBergen.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from friskby_controlpanel import friskby_controlpanel as cp
from tests.fake_friskby_interface import FakeFriskbyInterface
import unittest


class DeviceSetupTestCase(unittest.TestCase):

    def setUp(self):
        cp.app.config['FRISKBY_INTERFACE'] = FakeFriskbyInterface()
        cp.app.config['TESTING'] = True
        self.app = cp.app.test_client()

    def test_the_dash_unregistered_device(self):
        out = self.app.get('/', follow_redirects=True)
        self.assertIn('Not registered', out.data)

    def test_trying_to_register_nulldevice(self):
        out = self.app.post('/register', data=dict({
            "deviceid": ""
        }))
        self.assertIn('No device id', out.data)

    def test_successful_registering_of_device(self):
        out = self.app.post('/register', data=dict({
            "deviceid": "my-device-id"
        }), follow_redirects=True)
        self.assertIn('Device now registered', out.data)

    def test_registering_a_device_but_it_fails_miserably(self):
        cp.app.config['FRISKBY_INTERFACE'].fails = True
        out = self.app.post('/register', data=dict({
            "deviceid": "my-device-id"
        }), follow_redirects=True)
        self.assertIn('Error', out.data)

    def test_device_registered(self):
        cp.app.config['FRISKBY_INTERFACE'].device_id = 'mydevice'
        out = self.app.get('/')
        self.assertIn('mydevice', out.data)


if __name__ == '__main__':
    unittest.main()



from flask_wtf import FlaskForm
from wtforms import IntegerField, StringField
from wtforms.validators import DataRequired


class SettingsForm(FlaskForm):
    rpi_sample_time = IntegerField(
        'rpi_sample_time',
        validators=[DataRequired()]
    )
    rpi_control_panel_host = StringField(
        'rpi_control_panel_host',
        validators=[DataRequired()]
    )
    rpi_control_panel_port = IntegerField(
        'rpi_control_panel_port',
        validators=[DataRequired()]
    )
    rpi_sds011 = StringField(
        'rpi_sds011',
        validators=[DataRequired()]
    )



# -*- coding: utf-8 -*-
"""
    Friskby Interface

    In interface between Friskby Control panel, and everything required from
    the rest of the friskby universe.

    Note that this fake interface only considers the 'submitter' service to
    be worthy mocking. It also will only consider the 'restart' action.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime as dt


class FakeFriskbyInterface():

    def __init__(self):
        self.fails = False
        self.device_id = None

        self.uploaded_samples_count = 0
        self.all_samples_count = 0

        self.most_recently_uploaded = None
        self.most_recently_sampled = None

        self.sampler_status = None
        self.sampler_journal = []
        self.socket = None

        self.settings = dict()

    def _check_service(self, service_name):
        services = ['sampler', 'submitter', 'friskby', 'friskby_controlpanel']
        if service_name not in services:
            raise KeyError("No such service")

    def download_and_save_config(self, url, filename):
        if self.fails:
            raise ValueError("Was asked to fail for url %s -> %s." % (
                url, filename
            ))

    def get_device_id(self, _):
        return self.device_id

    def get_service_status(self, service):
        if service == 'sampler':
            return self.sampler_status

        self._check_service(service)

    def get_service_journal(self, service):
        if service == 'sampler':
            return self.sampler_journal
        else:
            return []

        self._check_service(service)

    def manage_service(self, service, action):
        self._check_service(service)

        if action != 'restart':
            raise KeyError('%s is not an action I know of...' % action)

    def get_uploaded_samples_count(self):
        return self.uploaded_samples_count

    def get_all_samples_count(self):
        return self.all_samples_count

    def get_most_recently_uploaded(self):
        return self.most_recently_uploaded

    def get_most_recently_sampled(self):
        return self.most_recently_sampled

    def get_socket_iface_name(self):
        return self.socket

    def get_recent_samples(self):
        limit = 10
        data = []
        for i in range(limit):
            row = (42+i, 42+i/10.0, 'PM10' if i % 2 else 'PM25', dt.now(), 0 if i < 2 else 1)
            data.append(row)
        return data

    def get_settings(self):
        return self.settings

    def set_settings(self, settings):
        self.settings = settings



# -*- coding: utf-8 -*-
"""
    (c) 2017 FriskbyBergen.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from friskby_controlpanel import friskby_controlpanel as cp
from tests.fake_friskby_interface import FakeFriskbyInterface
import unittest


class ServicesTestCase(unittest.TestCase):

    def setUp(self):
        self.iface = FakeFriskbyInterface()
        cp.app.config['TESTING'] = True
        cp.app.config['FRISKBY_INTERFACE'] = self.iface

        # Set device so that the dashboard does not redirect.
        self.iface.device_id = 'mydevice'

        self.app = cp.app.test_client()

    def test_that_header_reflects_sampler_status(self):
        self.iface.sampler_status = 'active'
        out = self.app.get('/')
        self.assertIn("Sampler <span>active</span>", out.data)

        self.iface.sampler_status = 'dead'
        out = self.app.get('/')
        self.assertIn("Sampler <span>dead</span>", out.data)

    def test_service_page(self):
        self.iface.sampler_status = 'superactive'
        self.iface.sampler_journal = [
            {
                "__REALTIME_TIMESTAMP": "1491213967810472",
                "__MONOTONIC_TIMESTAMP": "8281790",
                "MESSAGE_ID": "7d4958e842da4a758f6c1cdc7b36dcc5",
                "_PID": "1",
                "_COMM": "systemd",
                "_EXE": "/lib/systemd/systemd",
                "_CMDLINE": "/sbin/init splash",
                "UNIT": "rsyslog.service",
                "MESSAGE": "Starting System Logging Service...",
            }
        ]
        out = self.app.get('/service/sampler')
        self.assertIn(self.iface.sampler_status, out.data)
        self.assertIn("Starting System Logging Service..", out.data)

    def test_managing_non_existing_service(self):
        out = self.app.post('/service/baz/stop', follow_redirects=True)
        self.assertIn('No such service (baz) or action (stop).', out.data)

    def test_managing_non_service_with_bad_action(self):
        out = self.app.post('/service/sampler/fly_away', follow_redirects=True)
        self.assertIn('No such service (sampler) or action (fly_away).',
                      out.data)

    def test_manage_known_service(self):
        out = self.app.post('/service/sampler/restart', follow_redirects=True)
        self.assertIn('Requested restart on service sampler.',
                      out.data)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
    Parses journalctl json output.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
from datetime import datetime as dt
from datetime import timedelta as td


KEYS = {
    '__REALTIME_TIMESTAMP': 'timestamp_realtime',
    '__MONOTONIC_TIMESTAMP': 'timestamp_monotonic',
    'MESSAGE': 'message',
    'SYSLOG_IDENTIFIER': 'executable',
    '_SYSTEMD_UNIT': 'service'
}
PIDKEY = '_PID'


def from_timestamp(stamp, monotonic=False):
    int_stamp = int(stamp) // 1000000  # remove micro second
    if not monotonic:
        time_stamp = dt.fromtimestamp(int_stamp)
        iso_format = '%Y-%m-%d %H:%M:%S'
        return time_stamp.strftime(iso_format)
    else:
        delta = td(seconds=int_stamp)
        return str(delta)


def parse_message(k, v):
    if 'timestamp' in k:
        mono = 'monotonic' in k
        return from_timestamp(v, monotonic=mono)
    return v



# -*- coding: utf-8 -*-
"""
    (c) 2017 FriskbyBergen.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import fnmatch
import os
import unittest
from pylint import epylint as lint


class CodeTestCase(unittest.TestCase):

    def _get_lintable_files(self, folder):
        matches = []
        for root, _, filenames in os.walk(folder):
            for filename in fnmatch.filter(filenames, '*.py'):
                matches.append(os.path.join(root, filename))
        return matches

    def test_linting(self):
        files = self._get_lintable_files('friskby_controlpanel')
        files = files + self._get_lintable_files('tests')
        files.append('bin/fby_init')
        for f in files:
            self.assertEqual(lint.lint(f), 0, 'linting required for %s' % f)


if __name__ == '__main__':
    unittest.main()



ROOT_URL           = 'https://friskby.herokuapp.com'
SENSOR_PATH        = '/sensor/api/device'
DEVICE_CONFIG_PATH = '/usr/local/friskby/etc/config.json'

def config_url(device_id):
    return ROOT_URL + "%s/%s/" % (SENSOR_PATH, device_id)



# -*- coding: utf-8 -*-
"""
    Friskby Interface

    In interface between Friskby Control panel, and everything required from
    the rest of the friskby universe.

    Licence
    =======
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

try:
    import dbus
except ImportError:
    raise ImportError('Please install python-dbus.')

import socket
import json
import subprocess
import sys
from friskby import DeviceConfig, FriskbyDao
from rpiparticle import fby_settings


SYSTEMD_NAME = 'org.freedesktop.systemd1'
SYSTEMD_OBJ = '/org/freedesktop/systemd1'
SYSTEMD_UNIT_IFACE = 'org.freedesktop.systemd1.Unit'
SYSTEMD_MANAGER_IFACE = 'org.freedesktop.systemd1.Manager'
DBUS_PROPERTIES_IFACE = 'org.freedesktop.DBus.Properties'


# Represents a systemd DBus proxy as needed by the control panel.
class SystemdDBus():

    def __init__(self):
        self.sysbus = dbus.SystemBus()
        systemd1 = self.sysbus.get_object(SYSTEMD_NAME, SYSTEMD_OBJ)
        self.manager = dbus.Interface(systemd1, SYSTEMD_MANAGER_IFACE)

    def get_unit_status(self, unit):
        """Returns a UnitStatus or None if unit not installed.

            type UnitStatus struct {
                Name, Description, LoadState, ActiveState, SubState, Followed,
                Path, JobId, JobType, JobPath
            }
        """
        active_units = self.manager.ListUnits()
        for ret in active_units:
            if ret[0] == unit:
                return ret
        return None

    def restart_unit(self, unit):
        self.manager.RestartUnit(unit, 'replace')

    def start_unit(self, unit):
        self.manager.StartUnit(unit, 'replace')

    def stop_unit(self, unit):
        self.manager.StopUnit(unit, 'replace')


# Glues DBus, python-friskby and other pertinent things together so as to
# hide implementation details from the control panel.
class FriskbyInterface():
    known_units = {
        'friskby': 'friskby.service',
        'sampler': 'friskby-sampler.service',
        'submitter': 'friskby-submitter.service',
        'friskby_controlpanel': 'friskby-controlpanel.service',
    }

    def __init__(self):
        self.systemd = SystemdDBus()
        self.dao = FriskbyDao(fby_settings.get_setting("rpi_db"))

    def _service_to_unit(self, service):
        """Returns a unit or None if the service wasn't pertinent to the
        friskby system."""
        return self.known_units.get(service, None)

    def download_and_save_config(self, url, filename):
        """Downloads config from url and saves it to the given filename."""
        config = DeviceConfig.download(url)
        config.save(filename=filename)

    def get_service_status(self, service):
        """Returns the unit status as defined by [1]. Only services pertinent
        to the friskby system will be considered.
        [1] https://www.freedesktop.org/wiki/Software/systemd/dbus/
        """
        unit = self._service_to_unit(service)

        if service is None:
            raise KeyError('%s is not a pertinent service.' % service)

        status = self.systemd.get_unit_status(unit)
        try:
            return status[3]
        except TypeError:
            return None

    def get_service_journal(self, service, limit=50):
        """Returns a list of dicts containing output from journalctl where unit
        is the given service. Only services pertinent to the friskby system
        will be considered.

        Limit is the maximum amount of log lines to be returned. None implies
        no limit.

        See https://www.freedesktop.org/software/systemd/man/systemd.journal-fields.html  # noqa
        for information on what the list of dicts contains.
        """
        unit = self._service_to_unit(service)

        if service is None:
            raise KeyError('%s is not a pertinent service.' % service)

        output = None
        lines = []
        args = [
            "journalctl",
            "--unit=%s" % unit,
            "--catalog",
            "--full",
            "-o", "json"
        ]

        if limit is None:
            args = args + ["--lines", "all"]
        else:
            args = args + ["--lines", str(limit)]

        try:
            output = subprocess.check_output(args)
        except subprocess.CalledProcessError as e:
            # This means we got a non-zero exit code from journalctl. We can
            # do naught but log.
            print("Failed to capture journalctl output for %s:"
                  " %s exited with %d.\nOutput:%s" % (unit, e.cmd,
                                                      int(e.returncode),
                                                      e.output))
            sys.stdout.flush()
        else:  # No process error, so let's take a look at the output
            for line in output.split("\n"):
                if line != "":
                    js = json.loads(line)
                    lines.append(js)

        lines.reverse()
        return lines

    def get_device_id(self, filename):
        """Returns the device id, or None."""
        config = DeviceConfig(filename)
        device_id = config.getDeviceID()
        if device_id == "" or device_id is None:
            return None
        return device_id

    def get_uploaded_samples_count(self):
        fetch_uploaded = True
        return self.dao.get_num_rows(fetch_uploaded)

    def get_all_samples_count(self):
        return self.dao.get_num_rows()

    def get_most_recently_uploaded(self):
        fetch_uploaded_entry = True
        return self.dao.last_entry(fetch_uploaded_entry)

    def get_most_recently_sampled(self):
        return self.dao.last_entry()

    def get_recent_samples(self):
        data = self.dao.get_recent_samples(limit=10)
        return data

    @staticmethod
    def get_socket_iface_name():
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("friskby.no", 80))
            sockname = s.getsockname()[0]
            s.close()
            return sockname
        except socket.error:
            return None

    def manage_service(self, service, action):
        unit = self._service_to_unit(service)

        if action == 'restart':
            self.systemd.restart_unit(unit)
        elif action == 'start':
            self.systemd.start_unit(unit)
        elif action == 'stop':
            self.systemd.stop_unit(unit)
        else:
            raise KeyError('%s is not an action I know of...' % action)

    def manage(self, action):
        for service in self.known_units.keys():
            self.manage_service(service, action)

    def get_settings(self):
        return fby_settings.get_settings()

    def set_settings(self, settings):
        fby_settings.set_setting('rpi_sample_time',
                                 settings['rpi_sample_time'])
        fby_settings.set_setting('rpi_control_panel_host',
                                 settings['rpi_control_panel_host'])
        fby_settings.set_setting('rpi_control_panel_port',
                                 settings['rpi_control_panel_port'])
        fby_settings.set_setting('rpi_sds011',
                                 settings['rpi_sds011'])
